import { LRUCache } from "./lru";
import { merge } from "./merge";
import { ReconnectingWebSocket } from "./reconnectingwebsocket";
export const ErrorNotConnected = "not connected";
export const ErrorMutationTimeout = "mutation timed out";
export const ErrorConnectionClosed = "connection closed";
function dataFromSubscription(subscription) {
    return Object.freeze({
        state: subscription.state,
        value: subscription.value,
        error: subscription.error,
    });
}
export class GraphQLError extends Error {
    constructor(message, requestId) {
        super(message);
        this.requestId = requestId;
        // Set the prototype explicitly.
        Object.setPrototypeOf(this, GraphQLError.prototype);
    }
}
class Connection {
    constructor(connectFunction) {
        this.nextId = 0;
        this.subscriptions = new Map();
        this.mutations = new Map();
        this.past = new LRUCache(100);
        this.initialRetryDelay = 1000;
        this.maxRetryDelay = 60000;
        this.mutationTimeoutMs = 10000;
        this.socket = new ReconnectingWebSocket(connectFunction, {
            onOpen: this.handleOpen.bind(this),
            onClose: this.handleClose.bind(this),
            onMessage: this.handleMessage.bind(this),
        });
        this.socket.reconnect();
    }
    makeId() {
        return (this.nextId++).toString();
    }
    subscribe({ query, variables, observer, }) {
        const id = this.makeId();
        const cached = this.past.find({
            query,
            variables,
        });
        const subscription = {
            state: cached ? "cached" : "loading",
            retryDelay: this.initialRetryDelay,
            retryHandle: undefined,
            query,
            variables,
            observer,
            value: cached,
            error: undefined,
        };
        this.subscriptions.set(id, subscription);
        if (this.socket.state === "connected") {
            this.send({
                id,
                type: "subscribe",
                message: { query, variables },
            });
        }
        return {
            close: () => {
                const sub = this.subscriptions.get(id);
                if (!sub) {
                    return;
                }
                if (sub.value !== undefined) {
                    this.past.add({ query, variables }, sub.value);
                }
                this.subscriptions.delete(id);
                if (this.socket.state === "connected") {
                    this.send({ id, type: "unsubscribe" });
                }
            },
            data: () => {
                return dataFromSubscription(subscription);
            },
        };
    }
    mutate({ query, variables, }) {
        const id = this.makeId();
        if (this.socket.state === "connected") {
            this.send({ id, type: "mutate", message: { query, variables } });
        }
        else {
            return Promise.reject(new Error(ErrorNotConnected));
        }
        return new Promise((resolve, reject) => {
            const mutation = {
                query,
                variables,
                timeout: setTimeout(() => mutation.reject(new Error(ErrorMutationTimeout)), this.mutationTimeoutMs),
                resolve,
                reject,
            };
            this.mutations.set(id, mutation);
        });
    }
    notify(subscription) {
        subscription.observer(dataFromSubscription(subscription));
    }
    retry(id) {
        const subscription = this.subscriptions.get(id);
        if (subscription === undefined) {
            return;
        }
        this.send({
            id,
            type: "subscribe",
            message: { query: subscription.query, variables: subscription.variables },
        });
    }
    open() {
        this.socket.reconnect();
    }
    close() {
        this.socket.close();
    }
    send(message) {
        this.socket.send(message);
    }
    handleOpen() {
        for (const [id, subscription] of this.subscriptions) {
            this.send({
                id,
                type: "subscribe",
                message: {
                    query: subscription.query,
                    variables: subscription.variables,
                },
            });
            if (subscription.retryHandle) {
                clearTimeout(subscription.retryHandle);
                subscription.retryHandle = undefined;
            }
        }
    }
    handleClose() {
        for (const [, subscription] of this.subscriptions) {
            if (subscription.state === "subscribed") {
                subscription.state = "cached";
                this.notify(subscription);
            }
        }
        for (const [, mutation] of this.mutations) {
            mutation.reject(new Error(ErrorConnectionClosed));
        }
        this.mutations.clear();
    }
    handleMessage(envelope) {
        let subscription;
        let mutation;
        switch (envelope.type) {
            case "update":
                subscription = this.subscriptions.get(envelope.id);
                if (subscription !== undefined) {
                    if (subscription.state !== "subscribed") {
                        subscription.state = "subscribed";
                        subscription.error = undefined;
                        subscription.retryDelay = this.initialRetryDelay;
                    }
                    subscription.value = merge(subscription.value, envelope.message);
                    this.notify(subscription);
                }
                break;
            case "result":
                mutation = this.mutations.get(envelope.id);
                if (mutation !== undefined) {
                    mutation.resolve(merge(null, envelope.message));
                    clearTimeout(mutation.timeout);
                    this.mutations.delete(envelope.id);
                }
                break;
            case "error":
                const requestId = envelope.metadata && envelope.metadata.requestId;
                const requestIdMessage = requestId ? `RequestID:\n${requestId}\n` : "";
                subscription = this.subscriptions.get(envelope.id);
                if (subscription !== undefined) {
                    console.error("Subscription failed.", requestIdMessage, "Query:\n", subscription.query, "\nVariables:\n", subscription.variables, "\nError:\n", envelope.message);
                    subscription.state = "error";
                    subscription.error = new GraphQLError(envelope.message, requestId);
                    subscription.retryHandle = setTimeout(() => this.retry(envelope.id), subscription.retryDelay);
                    subscription.retryDelay = Math.min(this.maxRetryDelay, subscription.retryDelay * 2);
                    this.notify(subscription);
                }
                mutation = this.mutations.get(envelope.id);
                if (mutation !== undefined) {
                    console.error("Mutation failed.", requestId, "Query:\n", mutation.query, "\nVariables:\n", mutation.variables, "\nError:\n", envelope.message);
                    mutation.reject(new GraphQLError(envelope.message, requestId));
                    clearTimeout(mutation.timeout);
                    this.mutations.delete(envelope.id);
                }
                break;
            default:
                break;
        }
    }
}
export { Connection };
