var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
// PingingWebSocket is a WebSocket you can rely on. It has timeouts for all
// operations, and performs its own pings to make sure the connection to the
// backend remains functioning.
export class PingingWebSocket {
    constructor(connectFunction, listener, options = {}) {
        // Initialize our settings and state.
        this.connectFunction = connectFunction;
        this.listener = listener;
        this.options = Object.assign({}, PingingWebSocket.defaultOptions, options);
        this.hasShutdown = false;
    }
    // send sends a message string to the server. If the WebSocket isn't open,
    // send fails silently.
    send(message) {
        if (this.socket && this.socket.readyState === WebSocket.OPEN) {
            try {
                this.socket.send(JSON.stringify(message));
            }
            catch (error) {
                console.log(`pingingwebsocket: this.socket.send failed: ${error}`);
            }
        }
    }
    // close closes the PingingWebSocket, calling onClose on the listener if the
    // WebSocket is still open. After close returns, no more functions on the
    // listener will be called.
    close() {
        this.shutdown(PingingWebSocket.reasonCloseCalled);
    }
    // readyState returns the current underlying WebSocket's readyState.
    get readyState() {
        if (this.hasShutdown) {
            return WebSocket.CLOSED;
        }
        else if (this.socket) {
            return this.socket.readyState;
        }
        else {
            // If we have no socket yet, we are still waiting for the connect
            // promise to complete.
            return WebSocket.CONNECTING;
        }
    }
    connect() {
        return __awaiter(this, void 0, void 0, function* () {
            console.log("pingingwebsocket: connecting");
            // Start the connection timeout.
            this.connectTimeout = setTimeout(() => this.shutdown(PingingWebSocket.reasonConnectionTimeout), this.options.connectionTimeoutMs);
            // Invoke the connectFunction.
            let socket;
            try {
                socket = yield this.connectFunction();
            }
            catch (error) {
                this.shutdown(`connectFunction: ${error}`);
                return;
            }
            // Catch a shutdown that happened while connectFunction was running.
            if (this.hasShutdown) {
                socket.close();
                return;
            }
            // Store the socket. If we need to close it later, shutdown will do so.
            this.socket = socket;
            // Register socket handlers. All handlers are no-ops once we have shutdown.
            socket.onopen = () => {
                if (this.hasShutdown) {
                    return;
                }
                console.log("pingingwebsocket: opened");
                // Clear the timeout now that the connection has opened.
                if (this.connectTimeout) {
                    clearTimeout(this.connectTimeout);
                    this.connectTimeout = undefined;
                }
                // Start pinging. We will send the first ping after pingIntervalMs.
                this.sendPingTimeout = setTimeout(() => this.sendPing(), this.options.pingIntervalMs);
                // Notify the listener.
                this.listener.onOpen();
            };
            socket.onerror = (e) => {
                if (this.hasShutdown) {
                    return;
                }
                // Shutdown on error.
                this.shutdown(`socket error: ${e}`);
            };
            socket.onclose = (e) => {
                if (this.hasShutdown) {
                    return;
                }
                // Shutdown on close.
                this.shutdown(`socket close: ${e.reason}`);
            };
            socket.onmessage = (e) => {
                if (this.hasShutdown) {
                    return;
                }
                // Parse the message.
                let envelope;
                try {
                    envelope = JSON.parse(e.data);
                }
                catch (error) {
                    this.shutdown(`socket message JSON.parse: ${error}`);
                    return;
                }
                // If it is a ping reply, clear the timeout and schedule a new ping.
                if (envelope.type === "echo") {
                    if (this.receivePingTimeout) {
                        clearTimeout(this.receivePingTimeout);
                        this.receivePingTimeout = undefined;
                    }
                    if (this.sendPingTimeout) {
                        clearTimeout(this.sendPingTimeout);
                        this.sendPingTimeout = undefined;
                    }
                    this.sendPingTimeout = setTimeout(() => this.sendPing(), this.options.pingIntervalMs);
                }
                // Notify the listener.
                this.listener.onMessage(envelope);
            };
        });
    }
    // sendPing sends a ping and sets up a timeout for the response.
    sendPing() {
        if (this.sendPingTimeout) {
            clearTimeout(this.sendPingTimeout);
            this.sendPingTimeout = undefined;
        }
        this.send({ type: "echo" });
        if (this.receivePingTimeout) {
            clearTimeout(this.receivePingTimeout);
            this.receivePingTimeout = undefined;
        }
        this.receivePingTimeout = setTimeout(() => this.shutdown(PingingWebSocket.reasonPingTimeout), this.options.pingTimeoutMs);
    }
    // shutdown shuts down the websocket, clears all timeouts and notifies the
    // listener.
    shutdown(reason) {
        // Ensure shutdown is idempotent.
        if (this.hasShutdown) {
            return;
        }
        this.hasShutdown = true;
        console.log(`pingingwebsocket: shutdown: ${reason}`);
        // Shut down the webssocket.
        if (this.socket) {
            try {
                this.socket.close();
            }
            catch (error) {
                console.log(`pingingwebsocket: this.socket.close failed: ${error}`);
            }
            this.socket = undefined;
        }
        // Clear all timeouts.
        if (this.connectTimeout) {
            clearTimeout(this.connectTimeout);
            this.connectTimeout = undefined;
        }
        if (this.sendPingTimeout) {
            clearTimeout(this.sendPingTimeout);
            this.sendPingTimeout = undefined;
        }
        if (this.receivePingTimeout) {
            clearTimeout(this.receivePingTimeout);
            this.receivePingTimeout = undefined;
        }
        // Notify the listener.
        this.listener.onClose(reason);
    }
}
// defaultOptions provides a reasonable set of default PingingWebSocket options.
PingingWebSocket.defaultOptions = Object.freeze({
    connectionTimeoutMs: 30000,
    pingIntervalMs: 30000,
    pingTimeoutMs: 30000,
});
PingingWebSocket.reasonPingTimeout = "ping timeout";
PingingWebSocket.reasonConnectionTimeout = "connection timeout";
PingingWebSocket.reasonCloseCalled = "close called";
export default PingingWebSocket;
